@@ -1,37 +1,49 @@
+''' opencos.export_helper: package used by command sim and synth,
+
+to handle common tasks with "exporting" a DEPS target. An exported target copies all
+source files (deps, reqs) and attempts to resolve included files so they are all relative
+to +incdir+.
+'''
+
 import os
+import shutil
 import json
 
 from opencos import util
 from opencos.util import debug, info, warning, error
+from opencos.utils.markup_helpers import yaml_safe_writer
+
+# pylint: disable=dangerous-default-value
 
 
-_include_iteration_max_depth = 128 # Depth to look for nested included files.
-_remove_DEPS_yml_defines = [
+SV_INCLUDE_ITERATION_MAX_DEPTH = 128 # Depth to look for nested included files.
+REMOVE_DEPS_YML_DEFINES = [
     'OC_SEED',
     'OC_ROOT',
 ]
 
 
-def json_paths_to_jsonl(json_file_paths:list, output_json_path:str,
-                        assert_json_types=[dict]) -> None:
+def json_paths_to_jsonl(
+        json_file_paths:list, output_json_path:str, assert_json_types=[dict]
+) -> None:
     '''Given a list of .json filepath strs, save a single .jsonl (newline separated json(s)).
 
     errors if one of json_file_paths content's type is not in assert_json_types
     (assert_json_types can be empty list to avoid check).
     '''
 
-    if len(json_file_paths) == 0:
+    if not json_file_paths:
         error(f'{json_file_paths=} cannot be empty list')
 
 
     output_json_dir = os.path.split(output_json_path)[0]
     util.safe_mkdir(output_json_dir)
 
-    with open(output_json_path, 'w') as outf:
+    with open(output_json_path, 'w', encoding='utf-8') as outf:
 
         # jsonl is every line of the file is a json.
         for json_file_path in json_file_paths:
-            with open(json_file_path) as f:
+            with open(json_file_path, encoding='utf-8') as f:
                 data = json.load(f)
                 if len(assert_json_types) > 0 and type(data) not in assert_json_types:
                     error(f'{json_file_path=} JSON data is not a Table (py dict) {type(data)=}')
@@ -40,8 +52,9 @@ def json_paths_to_jsonl(json_file_paths:list, output_json_path:str,
         info(f'Wrote {len(json_file_paths)} tests to {output_json_path=}')
 
 
-def json_paths_to_single_json(json_file_paths:list, output_json_path:str,
-                              assert_json_types=[dict]) -> None:
+def json_paths_to_single_json(
+        json_file_paths:list, output_json_path:str, assert_json_types=[dict]
+) -> None:
     '''Given a list of .json filepath strs, save a single .json with key 'tests' and a list.
 
     errors if one of json_file_paths content's type is not in assert_json_types
@@ -55,13 +68,13 @@ def json_paths_to_single_json(json_file_paths:list, output_json_path:str,
     output_json_dir = os.path.split(output_json_path)[0]
     util.safe_mkdir(output_json_dir)
 
-    with open(output_json_path, 'w') as outf:
+    with open(output_json_path, 'w', encoding='utf-8') as outf:
 
         out_json_data = {
-            'tests': list(),
+            'tests': [],
         }
         for json_file_path in json_file_paths:
-            with open(json_file_path) as f:
+            with open(json_file_path, encoding='utf-8') as f:
                 data = json.load(f)
                 if len(assert_json_types) > 0 and type(data) not in assert_json_types:
                     error(f'{json_file_path=} JSON data is not a Table (py dict) {type(data)=}')
@@ -71,35 +84,22 @@ def json_paths_to_single_json(json_file_paths:list, output_json_path:str,
         info(f'Wrote {len(json_file_paths)} tests {output_json_path=}')
 
 
-def find_sv_included_files_within_file(filename:str,
-                                       known_incdir_paths:list,
-                                       warnings:bool=True,
-                                       modify_files_and_save_to_path=None,
-                                       unmodified_files_link_to_path=None) -> list:
-    '''Given a filename (full path) and a list of known incdir paths, returns
-    a list of included files (full path).
+def traverse_sv_file_for_includes(filename: str) -> (dict, set):
+    '''Lazily parses a SV <filename> looking for `includes
 
-    (Optional) modify_files_and_save_to_path (str: directory/path) if you wish
-    to strip all path information on the `include "(path)" for example:
-      `include "foo.svh" -- no modifications
-      `include "../bar.svh" -- is modified to become `include "bar.svh"
-    (Optional) unmodified_files_link_to_path (str: directory/path) if you wish
-    to symlink unmodified files to this path.
+    Returns tuple:
+       - dict of modified_lines: {linenum (int): line (str, modified line value), ...}
+       - set of found included files
     '''
 
-    found_included_files = set()
-
     assert any(filename.endswith(x) for x in ['.v', '.sv', '.vh', '.svh']), \
         f'{filename=} does not have a supported extension, refusing to parse it'
     assert os.path.exists(filename), f'{filename=} does not exist'
 
-    modified_lines = dict() # {linenum (int): line (str, modified line value), ...}
+    found_included_files = set()
+    modified_lines = {}
 
-    filename_no_path = os.path.split(filename)[1]
-
-    debug(f'export_helper: {filename=} {modify_files_and_save_to_path=} {unmodified_files_link_to_path=}')
-
-    with open(filename) as f:
+    with open(filename, encoding='utf-8') as f:
 
         for linenum, line in enumerate(f.readlines()):
             line_modified = False
@@ -111,7 +111,7 @@ def find_sv_included_files_within_file(filename:str,
                 parts = line.split("//")
                 words = parts[0].split() # only use what's on the left of the comments
                 prev_word_is_tick_include = False
-                for iter,word in enumerate(words):
+                for i,word in enumerate(words):
                     word = word.rstrip('\n')
                     if word == '`include':
                         # don't print this word, wait until next word
@@ -124,11 +124,10 @@ def find_sv_included_files_within_file(filename:str,
 
                         # strip the path information and keep track that
                         # we would like to modify this line of filename
-                        if modify_files_and_save_to_path:
-                            include_fname_no_path = os.path.split(include_fname)[1]
-                            if include_fname != include_fname_no_path:
-                                words[iter] = '"' + include_fname_no_path + '"'
-                                line_modified = True
+                        _, include_fname_no_path = os.path.split(include_fname)
+                        if include_fname != include_fname_no_path:
+                            words[i] = '"' + include_fname_no_path + '"'
+                            line_modified = True
 
                         if include_fname not in found_included_files:
                             # this has path information, perhaps relative, perhaps absolute, or
@@ -138,33 +137,85 @@ def find_sv_included_files_within_file(filename:str,
                 if line_modified:
                     modified_lines[linenum] = ' '.join(words)
 
-    debug(f'export_helper: {filename=} {modified_lines=}')
+    return modified_lines, found_included_files
+
+
+def write_modified_lines(
+        src_filename: str,
+        modified_lines: dict,
+        modify_files_and_save_to_path: str = '',
+        unmodified_files_copy_to_path: str = ''
+) -> None:
+    '''Given a dict of modified lines, walk the src_filename contents and write modifications
+
+    to the dst_filename. The modified_lines dict is:
+       {linenum (int): line (str, modified line value), ...}
+    '''
+
+    debug(f'export_helper: {src_filename=} {modified_lines=}')
+
+    _, src_filename_no_path = os.path.split(src_filename)
+
     # Optionally write out modified files (flatten the path information
     # on `include "../bar.svh" )
-    if len(modified_lines) > 0 and modify_files_and_save_to_path:
-        dst = os.path.join(modify_files_and_save_to_path, filename_no_path)
+    if modified_lines and modify_files_and_save_to_path:
+        dst = os.path.join(modify_files_and_save_to_path, src_filename_no_path)
         if not os.path.exists(dst):
-            with open(filename) as f, open(dst, 'w') as outf:
+            with open(src_filename, encoding='utf-8') as f, \
+                 open(dst, 'w', encoding='utf-8') as outf:
                 for linenum, line in enumerate(f.readlines()):
                     if linenum in modified_lines:
                         new_line = modified_lines[linenum]
                         outf.write(new_line + '\n')
-                        debug(f'export_helper: Modified {filename=} as {dst=}: {linenum=} {new_line=}')
+                        debug(f'export_helper: Modified {src_filename=} as {dst=}:',
+                              f'{linenum=} {new_line=}')
                     else:
                         outf.write(line)
 
-    # Optionally soft-link unmodified files to some path?
-    if len(modified_lines) == 0 and unmodified_files_link_to_path:
-        if os.path.isdir(unmodified_files_link_to_path):
-            dst = os.path.join(unmodified_files_link_to_path, filename_no_path)
+    # Copy unmodified files to some path.
+    if not modified_lines and unmodified_files_copy_to_path:
+        if os.path.isdir(unmodified_files_copy_to_path):
+            dst = os.path.join(unmodified_files_copy_to_path, src_filename_no_path)
             if not os.path.exists(dst):
-                debug(f'export_helper: Linked {filename=} to {dst=}')
-                os.symlink(src=filename, dst=dst)
+                debug(f'export_helper: Copied unmodified {src_filename=} to {dst=}')
+                shutil.copy(src=src_filename, dst=dst)
 
 
+def find_sv_included_files_within_file(
+        filename: str,
+        known_incdir_paths: list,
+        warnings: bool = True,
+        modify_files_and_save_to_path: str = '',
+        unmodified_files_copy_to_path: str = ''
+) -> list:
+    '''Given a filename (full path) and a list of known incdir paths, returns
+    a list of included files (full path).
+
+    (Optional) modify_files_and_save_to_path (str: directory/path) if you wish
+    to strip all path information on the `include "(path)" for example:
+      `include "foo.svh" -- no modifications
+      `include "../bar.svh" -- is modified to become `include "bar.svh"
+    (Optional) unmodified_files_copy_to_path (str: directory/path) if you wish
+    to copy unmodified files to this path.
+    '''
+
+    modified_lines, found_included_files = traverse_sv_file_for_includes(filename)
+
+    debug(f'export_helper: {filename=} {modify_files_and_save_to_path=}',
+          f'{unmodified_files_copy_to_path=}')
+
+    if modify_files_and_save_to_path or unmodified_files_copy_to_path:
+        # Save outputs to these paths:
+        write_modified_lines(
+            src_filename=filename,
+            modified_lines=modified_lines,
+            modify_files_and_save_to_path=modify_files_and_save_to_path,
+            unmodified_files_copy_to_path=unmodified_files_copy_to_path
+        )
+
     # Back to the list found_included_files that we observed within our filename, we
     # still need to return all the included files.
-    ret = list()
+    ret = []
     for fname in found_included_files:
         # Does this file exist, using our known_incdir_paths?
         found = False
@@ -174,7 +225,8 @@ def find_sv_included_files_within_file(filename:str,
                 if try_file_path not in ret:
                     ret.append(try_file_path)
                     found = True
-                    debug(f'export_helper: Include observed in {filename=} will use {try_file_path=} for export')
+                    debug(f'export_helper: Include observed in {filename=} will use',
+                          f'{try_file_path=} for export')
                     break # we can only match one possible file out of N possible incdir paths.
 
 
@@ -182,18 +234,20 @@ def find_sv_included_files_within_file(filename:str,
             # file doesn't exist in any included directory, we only warn here b/c
             # it will eventually fail compile.
             include_fname = fname
-            warning( f'export_helper: {include_fname=} does not exist in any of {known_incdir_paths=},' \
-                     + f'was included within source files: {filename=}'
-                    )
+            warning(f'export_helper: {include_fname=} does not exist in any of'
+                    f'{known_incdir_paths=}, was included within source files: {filename=}')
 
     return ret
 
 
-
-def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, target:str='',
-                               warnings:bool=True,
-                               modify_files_and_save_to_path=None,
-                               unmodified_files_link_to_path=None) -> list:
+def get_list_sv_included_files(
+        all_src_files: list,
+        known_incdir_paths: list,
+        target: str = '',
+        warnings: bool = True,
+        modify_files_and_save_to_path: str = '',
+        unmodified_files_copy_to_path: str = ''
+) -> list:
     ''' Given a list of all_src_files, and list of known_incdir_paths, returns a list
     of all included files (fullpath). This is recurisve if an included file includes another file.
 
@@ -205,14 +259,14 @@ def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, targ
         `include "foo.svh" -- no modifications
         `include "../bar.svh" -- is modified to become `include "bar.svh"
       Set to None (default) to disable.
-    unmodified_files_link_to_path -- (str: directory/path) if you wish to symlink unmodified
+    unmodified_files_copy_to_path -- (str: directory/path) if you wish to copy unmodified
       files to this path. Set to None (default) to disable.
     '''
 
     # order shouldn't matter, these will get added to the testrunner's filelist and
     # be included with +incdir+.
 
-    sv_included_files_dict = dict() # key, value is if we've traversed it (bool)
+    sv_included_files_dict = {} # key, value is if we've traversed it (bool)
 
     for fname in all_src_files:
         included_files_list = find_sv_included_files_within_file(
@@ -220,14 +274,14 @@ def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, targ
             known_incdir_paths=known_incdir_paths,
             warnings=warnings,
             modify_files_and_save_to_path=modify_files_and_save_to_path,
-            unmodified_files_link_to_path=unmodified_files_link_to_path
+            unmodified_files_copy_to_path=unmodified_files_copy_to_path
         )
 
         for f in included_files_list:
             if f not in sv_included_files_dict:
                 sv_included_files_dict[f] = False # add entry, mark it not traversed.
 
-    for _ in range(_include_iteration_max_depth):
+    for _ in range(SV_INCLUDE_ITERATION_MAX_DEPTH):
         # do these for a a depth of recurisve levels, in case `include'd file includes another file.
         # If we have more than N levels of `include hunting, then rethink this.
         # For example, some codebases would do their file dependencies as `include
@@ -239,7 +293,7 @@ def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, targ
                     known_incdir_paths=known_incdir_paths,
                     warnings=warnings,
                     modify_files_and_save_to_path=modify_files_and_save_to_path,
-                    unmodified_files_link_to_path=unmodified_files_link_to_path
+                    unmodified_files_copy_to_path=unmodified_files_copy_to_path
                 )
                 sv_included_files_dict[fname] = True # mark as traversed.
 
@@ -250,11 +304,11 @@ def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, targ
     if not all(sv_included_files_dict.values()):
         # we had some that we're traversed.
         not_traversed = [k for k,v in sv_included_files_dict.items() if not v]
-        error(f'Depth {_include_iteration_max_depth=} exceeded in looking for `includes,' \
+        error(f'Depth {SV_INCLUDE_ITERATION_MAX_DEPTH=} exceeded in looking for `includes,' \
               + f' {target=} {not_traversed=}')
 
 
-    ret = list()
+    ret = []
     for fname,traversed in sv_included_files_dict.items():
         if traversed:
             # add all the included files (should be traversed!) to our return list
@@ -264,6 +318,11 @@ def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, targ
 
 
 class ExportHelper:
+    '''ExportHelper is an object that command handlers can use to assist in creating
+
+    a directory with all exported sources, args, incdirs, defines, and output of
+    what was exported.
+    '''
 
     def __init__(self, cmd_design_obj, eda_command='export', out_dir=None, target=''):
         self.cmd_design_obj = cmd_design_obj
@@ -272,30 +331,40 @@ class ExportHelper:
         self.target = target
 
         self.args = self.cmd_design_obj.args # lazy alias.
-        self.included_files = list()
+        self.included_files = []
         self.out_deps_file = None
 
-        # TODO(drew) It would be neat if I could export an "eda multi" command, like
-        # CommandMulti that only gave me the list of all targets from a wildcard?
-        # Because then I could create exports for each individual target, but lump
-        # all files together and have a single exported DEPS.yml with unique targets.
 
-    def do_it(self, check_if_overwrite:bool=False,
-              deps_file_args:list=list(),
-              test_json_eda_config:dict=dict(), **kwargs):
+    def run(
+            self, check_if_overwrite:bool=False,
+            deps_file_args:list=[],
+            export_json_eda_config:dict={}, **kwargs
+    ) -> None:
+        '''main entrypoint for ExportHelper object. Creates output directory, writes files
 
-        self.make_out_dir(check_if_overwrite)
+        to it, creates a DEPS.yml in output directory, and optional output JSON file
+        '''
+
+        self.create_out_dir(check_if_overwrite)
         self.write_files_to_out_dir()
         self.create_deps_yml_in_out_dir(deps_file_args=deps_file_args)
 
-        if self.args.get('export-test-json', False):
-            self.create_test_json_in_out_dir(eda_config=test_json_eda_config, **kwargs)
+        if self.args.get('export-json', False):
+            self.create_export_json_in_out_dir(eda_config=export_json_eda_config, **kwargs)
 
         info(f'export_helper: done - wrote to: {self.out_dir}')
 
 
-    def make_out_dir(self, check_if_overwrite:bool=False):
-        assert self.args.get('top', ''), f'Need "top" to be set'
+    def create_out_dir(self, check_if_overwrite: bool= False) -> None:
+        '''Creates output directory for exported files, requires a 'top' to be
+
+        set by the original target, or inferred from target or files.
+        '''
+
+        if not self.args.get('top', ''):
+            error('export_helper.py internal error, args[top] is not set, cannot create',
+                  'output directory for export',
+                  f'{self.args=} {self.target=} {self.eda_command=} {self.out_dir=}')
 
         if not self.out_dir:
             if self.args.get('output', '') == "":
@@ -303,13 +372,20 @@ class ExportHelper:
 
         if check_if_overwrite and self.args.get('force', False):
             if os.path.exists(self.out_dir):
-                util.error(f"export_helper: output directory {out_dir} exists, use --force to overwrite")
+                error(f"export_helper: output directory {self.out_dir} exists, use --force",
+                      "to overwrite")
 
         if not os.path.exists(self.out_dir):
             info(f"export_helper: Creating {self.out_dir} for exported file tree")
             util.safe_mkdir(self.out_dir)
 
+
     def write_files_to_out_dir(self):
+        '''Called by self.run(), writes all files to output directory. Has to determine
+
+        the includes files used by SV/Verilog to unravel nested or relative included
+        paths.
+        '''
 
         # Also sets our list of included files.
         self.included_files = get_list_sv_included_files(
@@ -317,11 +393,19 @@ class ExportHelper:
             known_incdir_paths=self.cmd_design_obj.incdirs,
             target=self.target,
             modify_files_and_save_to_path=self.out_dir,
-            unmodified_files_link_to_path=self.out_dir
+            unmodified_files_copy_to_path=self.out_dir
         )
+
         info(f"export_helper: {self.target=} included files {self.included_files=}")
 
-    def create_deps_yml_in_out_dir(self, deps_file_args:list=list()):
+        for filename in self.cmd_design_obj.files_non_source:
+            dst = os.path.join(self.out_dir, os.path.split(filename)[1])
+            if not os.path.exists(dst):
+                shutil.copy(src=filename, dst=dst)
+
+
+    def create_deps_yml_in_out_dir(self, deps_file_args:list=[]):
+        '''Creates ouput exported directory DEPS.yml file with the exported target'''
         if not self.target:
             self.target = 'test'
         else:
@@ -332,7 +416,7 @@ class ExportHelper:
         info(f'export_helper: Creating DEPS.yml for {self.target=} in {self.out_dir=}')
 
         # Need to strip path information from our files_sv and files_v:
-        deps_files = list()
+        deps_files = []
         for fullpath in self.cmd_design_obj.files_sv + self.cmd_design_obj.files_v:
             filename = os.path.split(fullpath)[1]
             deps_files.append(filename)
@@ -340,36 +424,46 @@ class ExportHelper:
 
         data = {
             self.target: {
-                'incdirs': '.',
+                'incdirs': ['.'],
                 'deps': deps_files,
             }
         }
 
-        if len(deps_file_args) > 0:
+
+        if deps_file_args:
             data[self.target]['args'] = deps_file_args.copy()
 
         if self.args.get('top', None):
             data[self.target]['top'] = self.args['top']
 
-        if len(self.cmd_design_obj.defines.keys()) > 0:
+        if self.cmd_design_obj.defines:
             data[self.target]['defines'] = self.cmd_design_obj.defines.copy()
-            for define in _remove_DEPS_yml_defines:
+            for define in REMOVE_DEPS_YML_DEFINES:
                 # Remove defines keys for OC_ROOT and OC_SEED. Change OC_SEED to _ORIG_OC_SEED
                 if define in data[self.target]['defines']:
                     data[self.target]['defines'].pop(define)
 
+        reqs_fullpath_list = self.included_files + self.cmd_design_obj.files_non_source
+        if reqs_fullpath_list:
+            # Need to strip path information from non-source files:
+            data[self.target]['reqs'] = []
+            for fullpath in reqs_fullpath_list:
+                filename = os.path.split(fullpath)[1]
+                data[self.target]['reqs'].append(filename)
+
+
         dst = os.path.join(self.out_dir, 'DEPS.yml')
         self.out_deps_file = dst
-        util.yaml_safe_writer(data=data, filepath=dst)
+        yaml_safe_writer(data=data, filepath=dst)
 
 
-    def create_test_json_in_out_dir(self, eda_config:dict=dict(), **kwargs):
+    def create_export_json_in_out_dir( # pylint: disable=unused-argument
+            self, eda_config:dict={}, **kwargs
+    ) -> None:
+        '''Optionally creates an exported JSON file in the output directory'''
 
-        # We aren't going to do this if the eda_command was 'export':
         if not self.eda_command:
             return
-        if self.eda_command == 'export':
-            return
 
         # assumes we've run self.create_deps_yml_in_out_dir():
         assert self.target
@@ -382,17 +476,17 @@ class ExportHelper:
                 'multi': False, # Not yet implemented.
                 'command': self.eda_command,
                 'targets': [self.target],
-                'args': list(),
+                'args': [],
                 'waves': self.args.get('waves', False),
                 # tool - eda.CommandSimVerilator has this set in self.args:
                 'tool': self.args.get('tool', None),
             },
-            'files': list(),
+            'files': [],
         }
 
         # allow caller to override eda - tool, or eda - args, etc.
         for k,v in eda_config.items():
-            if k in data['eda'].keys() and v is not None:
+            if k in data['eda'] and v is not None:
                 data['eda'][k] = v
 
         # Note that args may already be set via:
@@ -402,7 +496,12 @@ class ExportHelper:
 
 
         all_files = [self.out_deps_file] + self.included_files \
-            + self.cmd_design_obj.files_sv + self.cmd_design_obj.files_v
+            + self.cmd_design_obj.files_sv + self.cmd_design_obj.files_v \
+            + self.cmd_design_obj.files_vhd + self.cmd_design_obj.files_cpp \
+
+        for x in self.cmd_design_obj.files_non_source:
+            if x not in all_files:
+                all_files.append(x)
 
         for somefile in all_files:
 
@@ -415,10 +514,10 @@ class ExportHelper:
             if os.path.exists(out_dir_filename):
                 somefile = out_dir_filename
             else:
-                raise(f'export test.json: {self.target=} Missing exported file, orig: {somefile=}')
+                error(f'export.json: {self.target=} Missing exported file, orig: {somefile=}')
 
             assert os.path.exists(somefile)
-            with open(somefile) as f:
+            with open(somefile, encoding='utf-8') as f:
                 filestr = ''.join(f.readlines())
                 data['files'].append({
                     'name': os.path.split(somefile)[1],
@@ -426,8 +525,18 @@ class ExportHelper:
                 })
 
 
-        dst = os.path.join(self.out_dir, 'test.json')
-        with open(dst, 'w') as f:
+        dst = os.path.join(self.out_dir, 'export.json')
+        with open(dst, 'w', encoding='utf-8') as f:
             json.dump(data, f)
             f.write('\n')
         info(f'export_helper: Wrote {dst=}')
+
+        # If this was from an `export` command, and the self.out_dir != self.args['work-dir'], then
+        # copy the export.json to the work-dir:
+        if self.out_dir != self.args['work-dir']:
+            util.safe_mkdirs(base=self.args['work-dir'], new_dirs=['export'])
+            src = os.path.abspath(os.path.join(self.out_dir, 'export.json'))
+            dst = os.path.join(self.args['work-dir'], 'export', 'export.json')
+            if not os.path.exists(dst):
+                shutil.copy(src=filename, dst=dst)
+            info(f'export_helper: Copied {src=} to {dst=}')
