@@ -1,7 +1,3 @@
-#!/usr/bin/python3
-
-# SPDX-License-Identifier: MPL-2.0
-
 ''' names.py
 
 This is a human-maintained but ideally machine parsable mapping of names
@@ -31,6 +27,7 @@ table = {
         3: "U50",
         4: "U55N",
         5: "U50C",
+        6: "PYNQ-Z2",
     },
 
     'OC_LIBRARY': {
